==> scribe-0.1.0-20180308.214329-67.jar: scribe/decimal.clj <==
(ns scribe.decimal
  "Utils for the conversions and massaging of BigDecimal values."
  (:require [abracad.avro :as aa]
            [scribe.time.utils :as u]
            [clojure.java.io :as io]
            [clojure.spec.alpha :as s])
  (:import java.nio.ByteBuffer
           [org.apache.avro Conversions$DecimalConversion LogicalTypes]))

(defn set-scale
  "Sets the scale of a BigDecimal."
  [^BigDecimal bd scale]
  (.setScale bd scale))

(defn set-precision
  "Sets the precision of a BigDecimal."
  [^BigDecimal bd precision]
  (let [prec (+ (.scale bd)
                (- precision (.precision bd)))]
    (.setScale bd prec)))

(defn- set-precision-scale
  "Best attempt to set the precision and scale of a BigDecimal."
  [bd precision scale]
  (-> bd
      (set-precision precision)
      (set-scale scale)))

(defn s-conform-decimal
  "Conforms a decimal? with precision and scale, returning
  :clojure.spec.alpha/invalid if unsuccessful.

  This is the `f` in the call to clojure.spec.alpha/conformer.

  NOTE you probably don't want to call this, use ->decimal-conformer instead."
  [precision scale bd]
  (if-not (decimal? bd)
    :clojure.spec.alpha/invalid
    (try
      (set-precision-scale bd precision scale)
      (catch ArithmeticException _
        :clojure.spec.alpha/invalid))))

(defmacro ->decimal-conformer
  "Clojure spec conformer, conforming a decimal? with precision and scale,
  returning :clojure.spec.alpha/invalid if unsuccessful."
  [& {:keys [precision scale]}]
  `(s/conformer (partial s-conform-decimal ~precision ~scale)))

(defn- ->LogicalType
  ^org.apache.avro.LogicalType
  [precision scale]
  (LogicalTypes/decimal (int precision) (int scale)))

(defn- ->avro-schema
  ^org.apache.avro.Schema
  [type precision scale]
  (aa/parse-schema
   {:type        type
    :logicalType :decimal
    :precision   precision
    :scale       scale}))

(defn decimal->bytes
  "Convert a BigDecimal to a ByteBuffer containing the two's-complement
  representation of its underlying unscaled BigInteger value in big-endian byte
  order."
  [bd precision scale]
  (.toBytes (Conversions$DecimalConversion.)
            (set-precision-scale bd precision scale)
            (->avro-schema :bytes precision scale)
            (->LogicalType precision scale)))

(defn decimal->fixed
  "Convert a BigDecimal to a org.apache.avro.generic.GenericFixed containing the two's-complement
  representation of its underlying unscaled BigInteger value in big-endian byte
  order.

  TODO FIXME not implemented"
  [bd precision scale]
  (u/not-implemented!)
  (.toFixed (Conversions$DecimalConversion.)
            (set-scale bd scale)
            (->avro-schema :fixed precision scale)
            (->LogicalType precision scale)))

(defn byte-buffer?
  "Is x an instance of a java.nio.ByteBuffer or any of its subclasses?"
  [x]
  (isa? (class x) java.nio.ByteBuffer))

(defn- ->buf
  "Half-assed cast to buffer if needed."
  ^ByteBuffer
  [buf-or-bytes]
  (if (byte-buffer? buf-or-bytes)
    buf-or-bytes
    (ByteBuffer/wrap buf-or-bytes)))

(defn bytes->decimal
  "Convert a byte array or ByteBuffer containing the two's-complement
  representation of an unscaled Integer value in big-endian byte
  order to a BigDecimal."
  [buffer-or-bytes precision scale]
  (.fromBytes (Conversions$DecimalConversion.)
              (->buf buffer-or-bytes)
              (->avro-schema :bytes precision scale)
              (->LogicalType precision scale)))

(defn fixed->decimal
  "Convert an org.apache.avro.generic.GenericFixed containing the
  two's-complement representation of an unscaled Integer value in big-endian
  byte order to a BigDecimal.

  TODO FIXME not really implemented"
  [^org.apache.avro.generic.GenericFixed fixed precision scale]
  (u/not-implemented!)
  (.fromFixed (Conversions$DecimalConversion.)
              fixed
              (->avro-schema :fixed precision scale)
              (->LogicalType precision scale)))

==> scribe-0.1.0-20180308.215744-69.jar: scribe/decimal.clj <==
(ns scribe.decimal
  "Utils for the conversions and massaging of BigDecimal values."
  (:require [abracad.avro :as aa]
            [scribe.time.utils :as u]
            [clojure.java.io :as io]
            [clojure.spec.alpha :as s])
  (:import java.nio.ByteBuffer
           [org.apache.avro Conversions$DecimalConversion LogicalTypes]))

(defn set-scale
  "Sets the scale of a BigDecimal."
  [^BigDecimal bd scale]
  (.setScale bd scale))

(defn set-precision
  "Sets the precision of a BigDecimal."
  [^BigDecimal bd precision]
  (let [prec (+ (.scale bd)
                (- precision (.precision bd)))]
    (.setScale bd prec)))

(defn- set-precision-scale
  "Best attempt to set the precision and scale of a BigDecimal."
  [bd precision scale]
  (-> bd
      (set-precision precision)
      (set-scale scale)))

(defn s-conform-decimal
  "Conforms a decimal? with precision and scale, returning
  :clojure.spec.alpha/invalid if unsuccessful.

  This is the `f` in the call to clojure.spec.alpha/conformer.

  NOTE you probably don't want to call this, use ->decimal-conformer instead."
  [precision scale bd]
  (if-not (decimal? bd)
    :clojure.spec.alpha/invalid
    (try
      (set-precision-scale bd precision scale)
      (catch ArithmeticException _
        :clojure.spec.alpha/invalid))))

(defmacro ->decimal-conformer
  "Clojure spec conformer, conforming a decimal? with precision and scale,
  returning :clojure.spec.alpha/invalid if unsuccessful."
  [& {:keys [precision scale]}]
  `(s/conformer (partial s-conform-decimal ~precision ~scale)))

(defn- ->LogicalType
  ^org.apache.avro.LogicalType
  [precision scale]
  (LogicalTypes/decimal (int precision) (int scale)))

(defn- ->avro-schema
  ^org.apache.avro.Schema
  [type precision scale]
  (aa/parse-schema
   {:type        type
    :logicalType :decimal
    :precision   precision
    :scale       scale}))

(defn decimal->bytes
  "Converts a BigDecimal to a ByteBuffer containing the two's-complement
  representation of its underlying unscaled BigInteger value in big-endian byte
  order."
  [bd precision scale]
  (.toBytes (Conversions$DecimalConversion.)
            (set-precision-scale bd precision scale)
            (->avro-schema :bytes precision scale)
            (->LogicalType precision scale)))

(defn decimal->fixed
  "Converts a BigDecimal to a org.apache.avro.generic.GenericFixed containing the two's-complement
  representation of its underlying unscaled BigInteger value in big-endian byte
  order.

  TODO FIXME not implemented"
  [bd precision scale]
  (u/not-implemented!)
  (.toFixed (Conversions$DecimalConversion.)
            (set-scale bd scale)
            (->avro-schema :fixed precision scale)
            (->LogicalType precision scale)))

(defn byte-buffer?
  "Is x an instance of a java.nio.ByteBuffer or any of its subclasses?"
  [x]
  (isa? (class x) java.nio.ByteBuffer))

(defn- ->buf
  "Half-assed cast to buffer if needed."
  ^ByteBuffer
  [buf-or-bytes]
  (if (byte-buffer? buf-or-bytes)
    buf-or-bytes
    (ByteBuffer/wrap buf-or-bytes)))

(defn bytes->decimal
  "Converts a byte array or ByteBuffer containing the two's-complement
  representation of an unscaled Integer value in big-endian byte
  order to a BigDecimal."
  [buffer-or-bytes precision scale]
  (.fromBytes (Conversions$DecimalConversion.)
              (->buf buffer-or-bytes)
              (->avro-schema :bytes precision scale)
              (->LogicalType precision scale)))

(defn fixed->decimal
  "Converts an org.apache.avro.generic.GenericFixed containing the
  two's-complement representation of an unscaled Integer value in big-endian
  byte order to a BigDecimal.

  TODO FIXME not really implemented"
  [^org.apache.avro.generic.GenericFixed fixed precision scale]
  (u/not-implemented!)
  (.fromFixed (Conversions$DecimalConversion.)
              fixed
              (->avro-schema :fixed precision scale)
              (->LogicalType precision scale)))
